/* ====================================================================
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2003 Matt Carter. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Apache" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */

#ifndef OAPR_COND_HPP
#define OAPR_COND_HPP

#include "oapr/oapr.hpp"
#include "oapr/buildconfig.hpp"
#include "oapr/time.hpp"
#include "oapr/mutex.hpp"
#include <prcvar.h>
#include <assert.h>

FILE_VERSION("$Revision: 1.5 $",oapr_cond_hpp);

namespace oapr
{
  class mutex;

  class LIBOAPR_API condition
  {
    public:
      condition();
      ~condition();
      /**
      * Put the active calling thread to sleep until signaled to wake up. Each
      * condition variable must be associated with a mutex, and that mutex must
      * be locked before calling this function, or the behavior will be
      * undefined. As the calling thread is put to sleep, the given mutex
      * will be simultaneously released; and as this thread wakes up the lock
      * is again simultaneously acquired.
      * @param cond the condition variable on which to block.
      * @param mutex the mutex that must be locked upon entering this function,
      *        is released while the thread is asleep, and is again acquired before
      *        returning from this function.
      */
      void wait(mutex& mtx);
      /**
       * Put the active calling thread to sleep until signaled to wake up or
       * the timeout is reached. Each condition variable must be associated
       * with a mutex, and that mutex must be locked before calling this
       * function, or the behavior will be undefined. As the calling thread
       * is put to sleep, the given mutex will be simultaneously released;
       * and as this thread wakes up the lock is again simultaneously acquired.
       * @param mutex the mutex that must be locked upon entering this function,
       *        is released while the thread is asleep, and is again acquired before
       *        returning from this function.
       * @param timeout The amount of time in microseconds to wait. This is 
       *        a maximum, not a minimum. If the condition is signaled, we 
       *        will wake up before this time, otherwise the error APR_TIMEUP
       *        is returned.
       * @retval OAPR_SUCCESS or OAPR_TIMEOUT.
       */
      oapr_status_t wait(mutex& mtx, const oapr_time_interval_t& timeout);

      /**
       * Signals all threads blocking on the given condition variable.
       * Each thread that was signaled is then schedule to wake up and acquire
       * the associated mutex. This will happen in a serialized manner.
       */
      void signal();
      
      /**\todo signal-waiting-thread signalling */
      
    private:
      /** pthreads condition variable */
    PRCondVar* m_cond;
  };
};

#endif
