/* ====================================================================
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2000-2003 Justin Kirby
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Apache" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 */
#ifndef OAPR_FILEBUF_HPP
#define OAPR_FILEBUF_HPP

#include "oapr/buildconfig.hpp"
#include <locale> // for std::codecvt
#include <streambuf>
#include "oapr/file.hpp"

FILE_VERSION("$Revision: 1.3 $",oapr_filebuf_hpp);

namespace oapr
{
  /**
   * list of policies to potentially add
   * - sync on dtor
   * - thread locks
   */
  template<typename charT,typename traits, std::streamsize bufszT=255>
  class basic_filebuf : public std::basic_streambuf<charT,traits>
  {
  public:
    // Types:
    typedef charT                     	        char_type;
    typedef traits                    	        traits_type;
    typedef typename traits_type::int_type	int_type;
    typedef typename traits_type::pos_type	pos_type;
    typedef typename traits_type::off_type	off_type;

    // Non-standard Types:
    typedef std::basic_streambuf<char_type, traits_type> streambuf_type;
    typedef basic_filebuf<char_type, traits_type>        filebuf_type;
    typedef typename traits_type::state_type             state_type;
    typedef std::codecvt<char_type, char,state_type>  codecvt_type;
    typedef typename codecvt_type::result 	         res_type;
    typedef std::ctype<char_type>                        ctype_type;

  protected:
    static const std::streamsize m_bufsz=bufszT;
    static const std::streamsize m_pbsize;//=bufszT/4;
    basic_file m_file;
    char_type m_inbuf[bufszT];/**< buffer used during input operations*/
    char_type m_outbuf[bufszT];/**< buffer used during output operations*/
  private:
    basic_filebuf(const basic_filebuf& fb);
    basic_filebuf& operator=(const basic_filebuf& fb);
  public:
    basic_filebuf()
      : m_file()
    {
      init_buffer();
    }

    virtual ~basic_filebuf()
    {
    }

    bool is_open()const{
      return m_file.is_open();
    }
    filebuf_type* open(const char* nm, std::ios_base::openmode md,
		       oapr_int32_t perms)
    {
      filebuf_type* ret=0;
      oapr_int32_t f=0;
      if(md&std::ios_base::in && md&std::ios_base::out)f|=PR_RDWR|PR_CREATE_FILE;
      else{
	if(md&std::ios_base::in) f |= PR_RDONLY;
	if(md&std::ios_base::out) f |= PR_WRONLY|PR_CREATE_FILE;
      }
      if(md&std::ios_base::app) f |= PR_APPEND;
      //      if(md&std::ios_base::binary) f |= APR_BINARY;
      if(md&std::ios_base::trunc) f |= PR_TRUNCATE;
      if(m_file.open(nm,f,perms)==OAPR_SUCCESS){
	ret=this;
      }
      return ret;
    }
    filebuf_type* close()
    {
      filebuf_type* ret=0;
      if(m_file.close()==OAPR_SUCCESS){
	ret=this;
      }
      return ret;
    }
  protected:
    /**
     * output
     */
    virtual int_type overflow(int_type c=traits_type::eof())
    {
      if(dump_buffer()<0){
	/**@todo throw or set failbit */
	return traits_type::eof();
      }else{
	if(!traits_type::eq_int_type(c,traits_type::eof())){
	  return sputc(c);
	}else{
	  return traits_type::not_eof(c);
	}
      }      
    }
    std::streamsize xsputn(const char_type* s, std::streamsize n)
    {
      if(n<epptr()-pptr()){
	std::memcpy(pptr(),s,n*sizeof(char_type));
	pbump(n);
	return n;
      }else{
	for(std::streamsize i=0;i<n;++i){
	  if(traits_type::eq_int_type(sputc(s[i]),traits_type::eof())){
	    return i;
	  }
	}
	return n;
      }
    }
    /**@todo implement following */
    /**
     * # seekoff
     * # seekpos
     * # imbue
     * # xsgetn
     */
    virtual int sync(){
      return dump_buffer();
    }

    /**
     * input
     */
    int_type underflow()
    {
      if(gptr()<egptr()){
	return traits_type::to_int_type(*gptr());
      }
      if(fill_buffer()<0){
	return traits_type::eof();
      }else{
	return traits_type::to_int_type(*gptr());
      }
    }
    int_type pbackfail(int_type c)
    {
      return 0;
    }
  private:
    void init_buffer()
    {
      static char_type* initpos=m_inbuf+filebuf_type::m_pbsize;
      setg(initpos,initpos,initpos);
      setp(m_outbuf,m_outbuf+bufszT);
    }

    int dump_buffer()
    {
      /**\todo the rv is actually bytes written not status*/
      int bs=pptr()-pbase();
      int rv=m_file.write(m_outbuf,bs);
      pbump(-bs);
      return rv;
    }
    int fill_buffer()
    {
      std::streamsize pb=std::min(gptr()-eback(),filebuf_type::m_pbsize);
      
      std::memmove(m_inbuf + (filebuf_type::m_pbsize-pb)*sizeof(char_type),
		   gptr()-pb*sizeof(char_type),
		   pb*sizeof(char_type));

      int rv=m_file.read(m_inbuf + (filebuf_type::m_pbsize*sizeof(char_type)),
			 filebuf_type::m_bufsz-filebuf_type::m_pbsize);
      if(rv<=0){
	setg(0,0,0);
	return -1;
      }else{
	setg(m_inbuf+filebuf_type::m_pbsize-pb,
	     m_inbuf+filebuf_type::m_pbsize,
	     m_inbuf+filebuf_type::m_pbsize+rv);
	return rv;
      }
    }


  };

  template<typename charT,typename traits, std::streamsize bufszT=255>
  const std::streamsize basic_filebuf<charT,traits,bufszT>::m_pbsize=bufszT/4;
};
#endif
