/* ====================================================================
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2003 Matt Carter. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Apache" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */

#ifndef OAPR_RWLOCK_HPP
#define OAPR_RWLOCK_HPP
 
#include "oapr/oapr.hpp"
#include "oapr/buildconfig.hpp"
#include <assert.h>
FILE_VERSION("$Revision: 1.4 $",rwlock_hpp);

namespace oapr
{
  class LIBOAPR_API rwlock
  {
    friend class rdlock;
    friend class wrlock;
  public:
    rwlock() { pthread_rwlock_init(&m_lock, NULL); }
    ~rwlock() { pthread_rwlock_destroy(&m_lock); }
  private:
    pthread_rwlock_t m_lock;
    rwlock(const rwlock&) {}
  };

  /**
   * An object protected by a rw lock.
   */
  class LIBOAPR_API rwobj
  {
    friend class rdlock;
    friend class wrlock;
    private:
      rwlock _lock;
  };

  /**
   * EXPERIMENTAL:
   * An object wrapped by a rw lock.
   */
  template<class T> class rw: public T, public rwobj
  {
  };
  
  /**
   * Scoped read lock. Holds a read lock on a rwlock, rwobj or rw<> object.
   */
  class LIBOAPR_API rdlock
  {
    public:
      explicit rdlock(rwlock& target): m_lock(&target.m_lock) { pthread_rwlock_rdlock(m_lock); }
      explicit rdlock(rwobj& target): m_lock(&target._lock.m_lock) { pthread_rwlock_rdlock(m_lock); }
      ~rdlock() { pthread_rwlock_unlock(m_lock); }
      /**
       * Cancel the rwlock before end of scope.
       */
      void cancel() { pthread_rwlock_unlock(m_lock); m_lock = NULL; }
      /**
       * Reacquire a cancelled rwlock before end of scope.
       */
      void reacquire(rwlock& target) { if(m_lock) cancel(); m_lock = &target.m_lock; pthread_rwlock_rdlock(m_lock); }
    private:
      pthread_rwlock_t *m_lock;

      rdlock() {}
      rdlock(const rdlock&) {}
  };

  /**
   * Scoped write lock. Holds a write lock on a rwlock, rwobj or rw<> object.
   */
  class LIBOAPR_API wrlock
  {
    public:
      explicit wrlock(rwlock& target): m_lock(&target.m_lock) { pthread_rwlock_wrlock(m_lock); }
      explicit wrlock(rwobj& target): m_lock(&target._lock.m_lock) { pthread_rwlock_wrlock(m_lock); }
      ~wrlock() { pthread_rwlock_unlock(m_lock); }
      /**
       * Cancel the rwlock before end of scope.
       */
      void cancel() { pthread_rwlock_unlock(m_lock); m_lock = NULL; }
      /**
       * Reacquire a cancelled rwlock before end of scope.
       */
      void reacquire(rwlock& target) { if(m_lock) cancel(); m_lock = &target.m_lock; pthread_rwlock_wrlock(m_lock); }
    private:
      pthread_rwlock_t *m_lock;

      wrlock() {}
      wrlock(const wrlock&) {}
  };

};

#endif
