/* ====================================================================
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2000-2003 Justin Kirby
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Apache" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 */
#include <sstream>
#include "oapr/buildconfig.hpp"
#include "oapr/socket.hpp"
#include "oapr/exception.hpp"

FILE_VERSION("$Revision: 1.6 $",oapr_io_socket_cpp);
namespace oapr
{
  oapr_status_t basic_socket::create(int family, int type)
  {
    if(is_created() || is_open()){
      OAPR_THROW_LVL(logic_error,
		     "socket is already initialized",exception_base::EXC_Error);
    }
    if(type == SOCK_DGRAM){
      m_sock = PR_OpenUDPSocket(family);
    }else if(type == SOCK_STREAM){
      m_sock = PR_OpenTCPSocket(family);
    }else{
      OAPR_THROW_LVL(logic_error,
		     "invalid type for socket",exception_base::EXC_Error);
    }
    if(m_sock){
      m_socktype=type;
      m_created=true;
      return OAPR_SUCCESS;
    } else {
      return OAPR_FAILURE;
    }

  }
  oapr_status_t basic_socket::shutdown(Shutdown_How how)
  {
    oapr_status_t rv = OAPR_FAILURE;
    if(is_open()){
      rv = PR_Shutdown(m_sock,(PRShutdownHow)how);
      if(rv == OAPR_SUCCESS){
        m_isopen=false;
        m_created=false;
      }
    }
    return rv;
  }
  oapr_status_t basic_socket::bind(const sockinfo& si)
  {
    if(!is_created()){
      OAPR_THROW_LVL(logic_error,
		     "socket is not created",exception_base::EXC_Error);
    }
    oapr_sockaddr_t sock=si.sockaddr();
    oapr_status_t stat = PR_Bind(m_sock,&sock);
    if(stat==OAPR_SUCCESS){
      m_isopen=true;
    }
    return stat;
  }
  oapr_status_t basic_socket::listen(oapr_int32_t backlog)
  {
    if(!is_open()){
      OAPR_THROW_LVL(logic_error,
		     "socket is not opened",exception_base::EXC_Crit);
    }
    return PR_Listen(m_sock,backlog);
  }
  basic_socket basic_socket::accept(oapr_time_interval_t timeout)
  {
    if(!is_open() || !is_created()){
      OAPR_THROW_LVL(logic_error,
		     "socket is not opened",exception_base::EXC_Crit);
    }
    oapr_sockaddr_t newaddr;
    oapr_socket_t *newsock = PR_Accept(m_sock,&newaddr,timeout);

    
    try{
      basic_socket sock(newsock,newaddr,m_socktype);
      return sock;
    }OAPR_CATCH()

  }
  oapr_status_t basic_socket::connect(const sockinfo &si,oapr_time_interval_t timeout)
  {
    if(!is_created()){
      OAPR_THROW_LVL(logic_error,"socket is not created",exception_base::EXC_Error);
    }
    oapr_sockaddr_t sock = si.sockaddr();
    oapr_status_t stat = PR_Connect(m_sock,&sock,timeout);
    if(stat==OAPR_SUCCESS){
      m_isopen=true;
    }
    return stat;
  }

  oapr_status_t basic_socket::set_option(Socket_Option opt, oapr_int32_t on)
  {
    if(!is_created()){
      OAPR_THROW_LVL(logic_error,
		     "socket is not created",exception_base::EXC_Error);
    }
    PRSocketOptionData data;
    //please forgive me for this switch
    switch(opt){
    case Opt_IpTtl:
      data.option = PR_SockOpt_IpTimeToLive;
      data.value.ip_ttl=on;
      break;
    case Opt_McastTtl:
      data.option = PR_SockOpt_McastTimeToLive;
      data.value.mcast_ttl = on;
      break;
    case Opt_Tos:
      data.option = PR_SockOpt_IpTypeOfService;
      data.value.tos = on;
      break;
    case Opt_NonBlocking:
      data.option = PR_SockOpt_Nonblocking;
      data.value.non_blocking = on;
      break;
    case Opt_ReuseAddr:
      data.option = PR_SockOpt_Reuseaddr;
      data.value.reuse_addr = on;
      break;
    case Opt_KeepAlive:
      data.option = PR_SockOpt_Keepalive;
      data.value.keep_alive = on;
      break;
    case Opt_McastLoopback:
      data.option = PR_SockOpt_McastLoopback;
      data.value.mcast_loopback = on;
      break;
    case Opt_NoDelay:
      data.option = PR_SockOpt_NoDelay;
      data.value.no_delay = on;
      break;
    case Opt_MaxSegment:
      OAPR_THROW(invalid_arg,"max segment is read only");
      break;
    case Opt_RecvBufferSize:
      data.option = PR_SockOpt_RecvBufferSize;
      data.value.recv_buffer_size = on;
      break;
    case Opt_SendBufferSize:     
      data.option = PR_SockOpt_SendBufferSize;
      data.value.send_buffer_size = on;
      break;
    default:
      OAPR_THROW(invalid_arg,"invalid option");
      break;
    }
    OAPR_CHECK(PR_SetSocketOption(m_sock,&data));
    return OAPR_SUCCESS;
  }

  oapr_status_t basic_socket::get_option(Socket_Option opt, oapr_int32_t& on)
  {
    if(!is_created()){
      OAPR_THROW_LVL(logic_error,
		     "socket is not created",exception_base::EXC_Error);
    }
    PRSocketOptionData data;
    //please forgive me for this switch
    switch(opt){
    case Opt_IpTtl:
      data.option = PR_SockOpt_IpTimeToLive;
      break;
    case Opt_McastTtl:
      data.option = PR_SockOpt_McastTimeToLive;
      break;
    case Opt_Tos:
      data.option = PR_SockOpt_IpTypeOfService;
      break;
    case Opt_NonBlocking:
      data.option = PR_SockOpt_Nonblocking;
      break;
    case Opt_ReuseAddr:
      data.option = PR_SockOpt_Reuseaddr;
      break;
    case Opt_KeepAlive:
      data.option = PR_SockOpt_Keepalive;
      break;
    case Opt_McastLoopback:
      data.option = PR_SockOpt_McastLoopback;
      break;
    case Opt_NoDelay:
      data.option = PR_SockOpt_NoDelay;
      break;
    case Opt_MaxSegment:
      data.option = PR_SockOpt_MaxSegment;
      break;
    case Opt_RecvBufferSize:
      data.option = PR_SockOpt_RecvBufferSize;
      break;
    case Opt_SendBufferSize:     
      data.option = PR_SockOpt_SendBufferSize;
      break;
    default:
      OAPR_THROW(invalid_arg,"invalid option");
      break;
    }
    OAPR_CHECK(PR_GetSocketOption(m_sock,&data));
    switch(opt){
    case Opt_IpTtl:
      on = data.value.ip_ttl;
      break;
    case Opt_McastTtl:
      on = data.value.mcast_ttl;
      break;
    case Opt_Tos:
      on = data.value.tos;
      break;
    case Opt_NonBlocking:
      on = data.value.non_blocking;
      break;
    case Opt_ReuseAddr:
      on = data.value.reuse_addr;
      break;
    case Opt_KeepAlive:
      on = data.value.keep_alive;
      break;
    case Opt_McastLoopback:
      on = data.value.mcast_loopback;
      break;
    case Opt_NoDelay:
      on = data.value.no_delay;
      break;
    case Opt_MaxSegment:
      on = data.value.max_segment;
      break;
    case Opt_RecvBufferSize:
      on = data.value.recv_buffer_size;
      break;
    case Opt_SendBufferSize:     
      on = data.value.send_buffer_size;
      break;
    default:
      OAPR_THROW(invalid_arg,"invalid option");
      break;
    }
    return OAPR_SUCCESS;
  }
  oapr_status_t basic_socket::close()
  {
    if(is_open()){
      oapr_status_t rv;
      rv = PR_Shutdown(m_sock,(PRShutdownHow)Shutdown_Both);
      m_isopen=false;
      m_created=false;
      return PR_Close(m_sock);
    }else{
      return OAPR_FAILURE;
    }
  }
  bool basic_socket::eof()const
  {
    return false;
  }

  oapr_int32_t basic_socket::write(const void* buff,oapr_int32_t toWrite)
  {
    if(!is_created() || !is_open()){
      OAPR_THROW_LVL(logic_error,
		     "socket is not open",exception_base::EXC_Error);
    }

    oapr_int32_t stat;
    switch(m_socktype){
    case 0:
      OAPR_THROW_LVL(invalid_arg,
		     "invalid socket type",exception_base::EXC_Error);
      break;
    case SOCK_STREAM:
      stat = PR_Send(m_sock,buff,toWrite,0,m_timeout);
      break;
    case SOCK_DGRAM:
      /**\todo check validity of m_remote*/
      stat = PR_SendTo(m_sock, buff,toWrite,0,&m_remote,m_timeout);
      break;
    default:
       OAPR_THROW_LVL(invalid_arg,
			   "unkown socket type",exception_base::EXC_Error);
       break;
    };
    //    OAPR_CHECK(stat);
    return stat;
  }
  oapr_int32_t basic_socket::read( void* buff,oapr_int32_t toRead)const
  {
    if(!is_created() || !is_open()){
      OAPR_THROW_LVL(logic_error,
		     "socket is not open",exception_base::EXC_Error);
    }
    oapr_status_t stat;
    switch(m_socktype){
    case 0:
      OAPR_THROW_LVL(invalid_arg,
		     "invalid socket type",exception_base::EXC_Error);
      break;
    case SOCK_STREAM:
      stat = PR_Recv(m_sock,buff,toRead,0,m_timeout);
      break;
    case SOCK_DGRAM:
      stat = PR_RecvFrom(m_sock,buff,toRead,0,&m_remote,m_timeout);
      break;
    default:
      OAPR_THROW_LVL(invalid_arg,
		     "unknown socket type",exception_base::EXC_Error);
    };
    if(stat == 0){
      /* network socket is closed */
      m_isopen=false;
    }

    return stat;
  }


  oapr_status_t basic_socket::flush()
  {
    return 0;
  }

  oapr_status_t basic_socket::timeout(oapr_time_interval_t timeout){
      m_timeout = timeout;
      return OAPR_SUCCESS;
    }
};
