/* ====================================================================
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2000-2003 Justin Kirby
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Apache" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 */
#ifndef OAPR_IO_SOCKET_HPP
#define OAPR_IO_SOCKET_HPP
#include "oapr/oapr.hpp"
#include "oapr/io.hpp"
#include "oapr/buildconfig.hpp"
#include "oapr/socketinfo.hpp"

FILE_VERSION("$Revision: 1.6 $",socket_hpp);

namespace oapr
{
  typedef PRFileDesc oapr_socket_t;
  enum Shutdown_How {
    Shutdown_Rcv = 0,
    Shutdown_Send = 1,
    Shutdown_Both = 2
  };
  enum Socket_Option{
    Opt_IpTtl=0,
    Opt_McastTtl,
    Opt_Tos,
    Opt_NonBlocking,
    Opt_ReuseAddr,
    Opt_KeepAlive,
    Opt_McastLoopback,
    Opt_NoDelay,
    Opt_MaxSegment,
    Opt_RecvBufferSize,
    Opt_SendBufferSize,
    /**@todo deal with these?
      PRLinger linger;
      PRMcastRequest add_member;
      PRMcastRequest drop_member; 
      PRNetAddr mcast_if;       
    */
  };
  /**
   * basic socket management class
   * Creates and maintians IO to a socket
   */
  class LIBOAPR_API basic_socket : public basic_io
  {
  public:
    basic_socket()
      : basic_io(),
	m_sock(0),
	m_socktype(0),
	m_created(false),
	m_isopen(false),
	m_timeout(PR_INTERVAL_NO_TIMEOUT)
    {}
    /**
     * Constructs a socket based from another one. Usually used
     * by accept.
     * @param sock The apr socket to assign to this
     * @param p The apr pool managing the socket
     * @param type The type of the socket.
     */
    basic_socket(oapr_socket_t *sock, oapr_sockaddr_t remote, oapr_int32_t type)
      : basic_io(),
	m_sock(sock),
	m_remote(remote),
	m_socktype(type),/**todo this is an error, need to get it*/
	m_created(true),
	m_isopen(true),
	m_timeout(PR_INTERVAL_NO_TIMEOUT)
    {}
    basic_socket(const basic_socket& s)
      : basic_io(s),
	m_sock(s.m_sock),
	m_remote(s.m_remote),
	m_local(s.m_local),
	m_socktype(s.m_socktype),
	m_created(s.m_created),
	m_isopen(s.m_isopen),
	m_timeout(s.m_timeout)
    {}
    virtual ~basic_socket(){}

    /**
     * Creates a socket
     * @param family The family 
     * @param type The type
     * @return result of the create
     */
    virtual oapr_status_t create(oapr_int32_t family, oapr_int32_t type);

    /**
     * Shutdown a socket cleanly
     * @param how How to shutdown the socket,
     * - OAPR_SHUTDOWN_READ no longer allow read requests
     * - OAPR_SHUTDOWN_WRITE no longer allow write requests
     * - OAPR_SHUTDOWN_READWRITE no longer allow read or writes requests
     * @return result of the shutdown
     */
    virtual oapr_status_t shutdown(Shutdown_How how);

    /**
     * Bind the socket to a set ip & port
     * @param sa created sockinfo
     * @return result of bind 
     * @throw oapr::logic_error if the socket is not created
     */
    virtual oapr_status_t bind(const sockinfo &sa);

    /**
     * Listen to a bound socket for connections
     * @param backlog Number of outstanding connections to allow
     * for the listen queue.
     * @throw oapr::logic_error is the socket is not open
     */
    virtual oapr_status_t listen(oapr_int32_t backlog);

    /**
     * Accept a new connection request
     * @return A new constructed socket
     * @throw oapr::logic_error if the socket is not created and opened
     * @throw oapr::oapr_exception if the pool for the new socket 
     * fails to create
     * @throw oapr::oapr_exception if the call to apr_socket_accept fails
     * @throw oapr::exception_base will be rethrown if the ctor of the
     * new socket throws
     */
    virtual basic_socket accept(oapr_time_interval_t timeout);

    /**
     * Issue a connection request to a socket.
     * @param si created sockinfo
     * @throw oapr::logic_error if the socket is not created
     */
    virtual oapr_status_t connect(const sockinfo &si,oapr_time_interval_t timeout);

    /**
     * Setup socket options 
     * @param opt Option to configure
     * @param on Value for the option
     * @throw oapr::logic_error if the socket is not created
     */
    virtual oapr_status_t set_option(Socket_Option opt, oapr_int32_t on);
    virtual oapr_status_t get_option(Socket_Option opt, oapr_int32_t& on);

    /**
     * Check whether the socket is created
     * @return true is the socket is created, false otherwise
     */
    virtual bool is_created()const{return m_created;}

    /**
     * Close the socket entirely
     * This performs an orderly shutdown of the socket and then
     * closes it
     * @return result of the close
     */
    virtual oapr_status_t close();

    /**
     * This is a no-op that will always return false
     */
    virtual bool eof()const;

    /**
     * Check the status of the socket
     * If the socket has been bound or connected this will return true
     */
    virtual bool is_open()const{return m_isopen;}

    /**
     * Write data out to socket
     * @param buff buffer to write
     * @param toWrite number of bytes to send
     * @param full wether to force sending of entire buffer.
     * @return number of bytes written
     * @throw oapr::logic_error if the socket is not open
     * @throw oapr::invalid_arg if the type is not of 
     * - SOCK_STREAM
     * - SOCK_DGRAM
     * @throw oapr::oapr_exception if the call to send fails 
     */
    virtual oapr_int32_t write(const void* buff, oapr_int32_t toWrite);

    /**
     * Read data from socket
     * @param buff buffer to read data into
     * @param toRead number of bytes to read
     * @param full whether to force fill the buffer
     * @return number of bytes read
     * @throw oapr::logic_error if the socket is not created and open
     * @throw oapr::invalid_arg if the the socket is not of
     * - SOCK_STREAM
     * - SOCK_DGRAM
     * @throw oapr::oapr_exception if the call to read fails
     */
    virtual oapr_int32_t read(void* buff,oapr_int32_t toRead)const;


    /**
     * currently a no-op
     */
    virtual oapr_status_t flush();

    /**
     * Setup socket timeout 
     * @param Value for the timeout
     * @return result of the set
     * @throw oapr::logic_error if the socket is not created
     */
    virtual oapr_status_t timeout(oapr_time_interval_t timeout);
    /**
     * GEt the timeout of the socket
     * @return the timeout of the socket
     * @throw oapr::logic_error if the socket is not created
     */
    virtual oapr_time_interval_t timeout()const{return m_timeout;}

  protected:
  private:
    oapr_socket_t *m_sock;/**< The actual socket*/
    mutable oapr_sockaddr_t m_remote;/**< remote socket addr*/
    oapr_sockaddr_t m_local;/**< local socket addr*/
    oapr_int32_t m_socktype;/**< the type of the socket*/
    bool m_created;/**< status flag whether the socket has been created*/
    mutable bool m_isopen;/**< whether the socket is connected/bound*/
    oapr_time_interval_t m_timeout;
  };
};
#endif
