/* ====================================================================
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2000-2003 Justin Kirby
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Apache" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 */
#ifndef OAPR_TIME_HPP
#define OAPR_TIME_HPP
#include "oapr-config.h"
#include "oapr/oapr.hpp"
#include "oapr/buildconfig.hpp"
#include "oapr/exception.hpp"
#include <time.h>
#include <string>
#include <ostream>

FILE_VERSION("$Revision: 1.8.2.3 $",oapr_time_hpp);

namespace oapr
{
  /**
   * Time location type. One of tl_utc for UTC, tl_local for Local Time.
   */
  enum time_location_type {
    UTC,
    LOCAL,
    ANY_LOCATION_TYPE
  };

  /**
   * Time location traits for UTC.
   */
  template<typename time_traits> struct utc_traits
  {
    typedef typename time_traits::time_type time_type;
    typedef typename time_traits::time_exp_t time_exp_t;
    inline time_location_type differentiator() { return UTC; }
    inline static void expand(time_type time, time_exp_t **exp) { time_traits::exp_expand_utc(time, exp); }
    inline static void compact(time_type *time, time_exp_t **exp) { time_traits::exp_compact_utc(time, exp); }
    static void assign_from_complement(const time_type& source, time_type *dest, time_exp_t **dest_exp)
    {
      // Convert local to utc time
      if(*dest_exp) time_traits::exp_free(dest_exp);
      time_traits::exp_expand_localtime(source, dest_exp);
      (*dest_exp)->tm_sec -= (*dest_exp)->tm_gmtoff;
      time_traits::exp_compact_utc(dest, dest_exp); // Leaves the (normalized) exp in place
    }
  };

  /**
   * Time location traits for Local Time.
   */
  template<typename time_traits> struct localtime_traits
  {
    typedef typename time_traits::time_type time_type;
    typedef typename time_traits::time_exp_t time_exp_t;
    inline time_location_type differentiator() { return LOCAL; }
    inline static void expand(time_type time, time_exp_t **exp) { time_traits::exp_expand_localtime(time, exp); }
    inline static void compact(time_type *time, time_exp_t **exp) { time_traits::exp_compact_localtime(time, exp); }
    static void assign_from_complement(const time_type& source, time_type *dest, time_exp_t **dest_exp)
    {
      // Convert utc to local time
      if(*dest_exp) time_traits::exp_free(dest_exp);
      time_traits::exp_expand_utc(source, dest_exp);
      (*dest_exp)->tm_sec += (*dest_exp)->tm_gmtoff;
      time_traits::exp_compact_localtime(dest, dest_exp); // Leaves the (normalized) exp in place
    }
  };

  /**
   * Basic time traits, 1 second resolution.
   * @remark Every standard system has time.h, these time traits are a C++ wrapper around those functions.
   * The types below must match the target platform.
   */
  struct basic_time_traits {
    /**
     * Composite time
     */
#if defined(HOST_OSX)
    typedef time_t time_type;
#else
    typedef __time_t time_type;
#endif
    /**
     * Expanded (broken-down) time
     */
    typedef struct ::tm time_exp_t;
    /**
     * Type used for gmt offset
     */
    typedef long gmt_offset_t;
    /**
     * Precision in 1/n seconds
     */
    inline int precision() { return 1; }
    /**
     * Return a new copy of the given expanded time structure.
     */
    inline static void exp_new_copy(time_exp_t *source, time_exp_t **newdest) {
      *newdest = new time_exp_t;
      memcpy(*newdest, source, sizeof(time_exp_t));
    }
    /**
     * Destroy the expanded time structure.
     * @remark Called only from destructors.
     */
    inline static void exp_destroy(time_exp_t **exp) {
      delete *exp;
    }
    /**
     * Free the expanded time structure and null the pointer.
     */
    inline static void exp_free(time_exp_t **exp) {
      delete *exp; *exp = NULL;
    }
    /**
     * Expand a composite time relative to Local Time into it's component parts.
     */
    static void exp_expand_localtime(time_type time, time_exp_t **exp) {
      *exp = new time_exp_t;
      if(!localtime_r(&time, *exp)) {
        exp_free(exp);
        OAPR_THROW_OAPR(OAPR_YEAR_OUT_OF_RANGE);
      }
    }
    /**
     * Expand a composite time relative to GMT (Universal Time Coordinated) into it's component parts.
     */
    static void exp_expand_utc(time_type time, time_exp_t **exp) {
      *exp = new time_exp_t;
      if(!gmtime_r(&time, *exp)) {
        exp_free(exp);
        OAPR_THROW_OAPR(OAPR_YEAR_OUT_OF_RANGE);
      }
    }
    /**
     * Compact an expanded time structure relative to local time into a composite representation.
     */
    static void exp_compact_localtime(time_type *time, time_exp_t **exp)
    {
      *time = ::mktime(*exp);
      if(*time == ((time_type)-1)) OAPR_THROW_OAPR(OAPR_MKTIME_FAILED);
    }
    /**
     * Compact an expanded time structure relative to GMT (Universal Time Coordinated) into a composite representation.
     */
    static void exp_compact_utc(time_type *time, time_exp_t **exp)
    {
      (*exp)->tm_isdst = -1/*Information not available*/; // Hmmm :)
      *time = ::mktime(*exp);
      if(*time == ((time_type)-1)) OAPR_THROW_OAPR(OAPR_MKTIME_FAILED);
    }
    /**
     * Return an RFC 822 representation of the given expanded time structure.
     */
    inline static std::string rfc822(time_exp_t *tm) {
      char buf[26];
      ::asctime_r(tm, buf);
      buf[24] = '\0';
      return std::string(buf);
    }
    /**
     * Return the current time as composite.
     */
    inline static time_type now() {
      time_type t;
      t = ::time(0);
      if(t == ((time_type)-1)) OAPR_THROW_OAPR(OAPR_GET_TIME_FAILED);
      return t;
    }
/*    inline void exp_assign(time_exp_t *source, time_exp_t **dest) {
      if(source) {
        if(*dest) delete *dest;
        *dest = new time_exp_t;
        memcpy(*dest, source, sizeof(time_exp_t));
      }
      else *dest = NULL;
    }*/
  };

  /**
   * Nanosecond resolution.
   */
  struct nano_time_traits {
    // TODO
  };

  /**
   * Simple time object.
   */
  template<typename traits, typename location_traits, typename complementary_location_traits> class basic_time
  {
  public:
//    typedef traits<time_location> _traits;
    typedef typename traits::gmt_offset_t gmt_offset_t;
    typedef typename traits::time_type time_type;
    typedef typename traits::time_exp_t time_exp_t;
    enum _initial_time_t { now };
    typedef _initial_time_t initial_time_t;
  public:
    /**
     * Create an empty time object.
     */
    basic_time(): m_time(0), m_exp(NULL) {}

    /**
     * Create an initialized time object.
     * @param init 'now' is the only value currently allowed.
     */
    basic_time(initial_time_t init): m_time(traits::now()), m_exp(NULL) {}

    /**
     * Copy constructor
     * @remark Doesn't copy the expanded data, to minimize baggage
     */
    basic_time(const basic_time& t): m_time(t.m_time), m_exp(NULL) {}

    /**
     * UTC<->Local Complementary copy constructor
     */
    explicit basic_time(const basic_time<traits, complementary_location_traits, location_traits>& t)
      : m_time(0), m_exp(NULL)
    {
      location_traits::assign_from_complement(t.m_time, &m_time, &m_exp);
    }

    /**
     * Create a time object from a native time value.
     * @param native_time Native time value to initialize from.
     */
    basic_time(time_t native_time): m_time(native_time), m_exp(NULL) {}

/*    time(int sec, int min, int hour, int day, int month, int year, bool is_dst = false)
      : m_time(0), m_exp(NULL)
    {
      // TODO. is_dst only applicable to local_time (specialization?)
    } */

    /**
     * Destructor.
     */
    virtual ~basic_time() { if(m_exp) traits::exp_destroy(&m_exp); }

    /**
     * Copy assignment operator.
     */
    basic_time& operator=(const basic_time& t) {
      if(this != &t/*self-copy*/) {
        m_time = t.m_time;
        if(m_exp) traits::exp_free(&m_exp);
        // traits::exp_assign(t.m_exp, &m_exp);
        return *this;
      }
    }

    /** Native assignment operator. Assign from the time_t of the platform. */
    basic_time& operator=(time_type native_time) {
      m_time = native_time;
      if(m_exp) traits::exp_free(&m_exp);
      return *this;
    }

    /**
     * UTC<->Local assignment operator
     */
    basic_time& operator=(const basic_time<traits, complementary_location_traits, location_traits>& t) {
      // Self copy not possible so not checked
      location_traits::assign_from_complement(t.m_time, &m_time, &m_exp);
    }

    friend class basic_time<traits, complementary_location_traits, location_traits>;

    /**
     * The  number of seconds after the minute, normally in the range 0
     * to 59, but can be up to 61 to allow for leap seconds.
     */
    int sec()    const { expand(); return m_exp->tm_sec; }
    /**
     * The number of minutes after the hour, in the range 0 to 59.
     */
    int min()    const { expand(); return m_exp->tm_min; }
    /**
     * The number of hours past midnight, in the range 0 to 23.
     */
    int hour()   const { expand(); return m_exp->tm_hour; }
    /**
     * The day of the month, in the range 1 to 31.
     */
    int day()    const { expand(); return m_exp->tm_mday; }
    /**
     * The number of months since January, in the range 0 to 11.
     */
    int month()  const { expand(); return m_exp->tm_mon; }
    /**
     * The number of years since 1900.
     */
    int year()   const { expand(); return m_exp->tm_year; }
    /**
     * The number of days since Sunday, in the range 0 to 6.
     */
    int wday()   const { expand(); return m_exp->tm_wday; }
    /**
     * The number of days since January 1, in the range 0 to 365.
     */
    int yday()   const { expand(); return m_exp->tm_yday; }
    /**
     * A flag that indicates whether daylight saving time is in  effect
     * at the time described.  The value is positive if daylight saving
     * time is in effect, zero if it is not, and negative if the information
     * is not available.
     */
    bool is_dst() const { expand(); return m_exp->tm_isdst; }
    /**
     * Seconds east of UTC (GMT).
     */
    gmt_offset_t gmt_offset() { expand(); return m_exp->tm_gmtoff; }

    /**
     * Formats dates in the RFC822 format.
     * Same as str().
     */
    /**\todo Check if its possible to alloc a new string and write to its internal buffer */
    std::string rfc822() const { expand(); return traits::rfc822(m_exp); }

    /**
     * Formats dates in the RFC822 / ctime() format.
     * Unlike ANSI/ISO C ctime(), oapr::time::str() does not include a \n at the end of the string.
     */
    std::string str() const { expand(); return traits::rfc822(m_exp); }

    friend std::ostream& operator<< <>(std::ostream&, const basic_time&);

  protected:
    /**
     * Expands the time into it's component parts, ready for accessing (e.g. sec(), wday())
     * @note Logically const because the represented time isn't affected by the expansion.
     */
    virtual void expand() const {
      if(!m_exp) location_traits::expand(m_time, &m_exp);
    }
    /**
     * Compacts the expanded time structure into a composite representation.
     */
/*    virtual void compact() {
      if(m_exp) {
        location_traits::compact(&m_time, &m_exp);
        // Delete what we just compacted
        delete exp; exp = NULL;
      }
    } NOT CURRENTLY USED or public (expanded time never changed) */
    /** native composite time */
    time_type m_time;
    /** Expanded time structure - allocated when used or after explicit expand() call. Freed if compact() is called. */
    mutable time_exp_t *m_exp;
  };

  template<typename traits, typename loc, typename cloc>
    std::ostream& operator<<(std::ostream& s, const basic_time<traits,loc,cloc>& t) { return s<<t.str(); }

  //  typedef basic_time<basic_time_traits<> > time;
  typedef basic_time<basic_time_traits, utc_traits<basic_time_traits>, localtime_traits<basic_time_traits> > utc_time;
  typedef basic_time<basic_time_traits, localtime_traits<basic_time_traits>, utc_traits<basic_time_traits> > local_time;
  typedef local_time time;

// TODO: var_time class which can be either local or utc (EXPERIMENTAL)
/*  template<typename traits> class var_time: public basic_time<basic_time_traits<> >
  {
  public:
    typedef typename var_time::initial_time_t initial_time_t;
    typedef basic_time<basic_time_traits<> > basic_time;
//    typedef time_location_type;
  public:
    var_time(): basic_time(), m_type(LOCAL) {}
    var_time(initial_time_t init): basic_time(init), m_type(LOCAL) {}
    var_time(const var_time& t): basic_time(t), m_type(t.m_type) {}
    var_time(const basic_time< basic_time_traits<LOCAL> >& t): basic_time(t.m_time), m_type(LOCAL) {}
    var_time(const basic_time< basic_time_traits<UTC> >& t): basic_time(t.m_time), m_type(UTC) {}
    var_time(time_t native_time, time_location_type time_location = LOCAL): basic_time(native_time), m_type(time_location) {}

    void set_type(time_location_type loc) { m_type = loc; }
    inline void local() { set_type(LOCAL); }
    inline void utc() { set_type(UTC); }
  protected:*/
    /**
     * Expands the time into it's component parts, ready for accessing (e.g. sec(), wday())
     * @note Logically const because time is not affected by expansion.
     */
/*  private:
    time_location_type m_type;
  };*/


  /**
   * Interval
   */
  typedef struct timespec native_interval_t;
  typedef __time_t second_t;
  typedef long int nanosecond_t;

  /**
   * Simple interval type
   */
  class interval {
  public:
    interval() { m_int.tv_sec = 0; m_int.tv_nsec = 0; }
    interval(second_t sec, nanosecond_t nsec = 0) { m_int.tv_sec = sec;  m_int.tv_nsec = nsec; }
    interval(const native_interval_t& native_interval) { m_int.tv_sec = native_interval.tv_sec;
                                                         m_int.tv_nsec = native_interval.tv_nsec; }
    interval(const interval& source) { m_int.tv_sec = source.m_int.tv_sec;
                                       m_int.tv_nsec = source.m_int.tv_nsec; }
    interval& operator=(const native_interval_t& native_interval) {  m_int.tv_sec = native_interval.tv_sec;
                                                        m_int.tv_nsec = native_interval.tv_nsec; return *this; }
    second_t sec() { return m_int.tv_sec; }
    void sec(second_t sec) { m_int.tv_sec = sec; }
    nanosecond_t nsec() { return m_int.tv_nsec; }
    void nsec(nanosecond_t nsec) { m_int.tv_nsec = nsec; }
  private:
    native_interval_t m_int;
  };




  /**
   * Returns true if the given year is a leap year.
   */
  bool is_leap_year(int year);

  /**
   * Returns the number of days in the given year.
   */
  int days_in_year(int year);

};
#endif
