/* ====================================================================
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2000-2003 Justin Kirby
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Apache" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 */
#ifndef OAPR_URI_HPP
#define OAPR_URI_HPP
#include <string>
#include <boost/spirit.hpp>
#include <boost/function.hpp>
namespace oapr
{
  class uri
  {
  public:
    uri()
      : m_isValid(false){}
    uri(const std::string& s)
      : m_isValid(false)
    {
      parse(s);
    }
    uri(const uri& u)
      : m_isValid(u.m_isValid),
	m_scheme(u.m_scheme),
	m_uri(u.m_uri){}
    virtual ~uri(){}

    virtual void parse(const std::string& u)
    {
      uri::uri_grammar g(this);
      m_isValid = boost::spirit::parse(u.c_str(),g).full;
      m_uri=u;
    }
    virtual std::string get_uri()const{return m_uri;}
    virtual bool is_valid()const{return m_isValid;}

    virtual std::string scheme()const{return m_scheme;}
    virtual void scheme(const std::string& s){m_scheme=s;}



  protected:
    virtual void on_query(char const* first,char const* last)=0;
    virtual void on_fragment(char const* first,char const* last)=0;
    virtual void on_param(char const* first,char const* last)=0;
    virtual void on_segment(char const* first,char const* last)=0;
    virtual void on_path_segments(char const* first,char const* last)=0;
    virtual void on_abs_path(char const* first,char const* last)=0;
    virtual void on_path(char const* first,char const* last)=0;
    virtual void on_port(char const* first,char const* last)=0;
    virtual void on_ipv4(char const* first,char const* last)=0;
    virtual void on_toplabel(char const* first,char const* last)=0;
    virtual void on_domainlabel(char const* first,char const* last)=0;
    virtual void on_hostname(char const* first,char const* last)=0;
    virtual void on_host(char const* first,char const* last)=0;
    virtual void on_hostport(char const* first,char const* last)=0;
    virtual void on_userinfo(char const* first,char const* last)=0;
    virtual void on_server(char const* first,char const* last)=0;
    virtual void on_reg_name(char const* first,char const* last)=0;
    virtual void on_authority(char const* first,char const* last)=0;
    virtual void on_scheme(char const* first,char const* last){
      m_scheme.assign(first,last);
    }
    virtual void on_rel_segment(char const* first,char const* last)=0;
    virtual void on_rel_path(char const* first,char const* last)=0;
    virtual void on_net_path(char const* first,char const* last)=0;
    virtual void on_opaque_part(char const* first,char const* last)=0;
    virtual void on_hier_part(char const* first,char const* last)=0;
    virtual void on_relative_uri(char const* first,char const* last)=0;
    virtual void on_absolute_uri(char const* first,char const* last)=0;
  private:
    bool m_isValid;
    /* <scheme>:<authority><path>?<query> */
    std::string m_scheme;
    std::string m_uri;
  protected:
    class uri_cb
    {
    public:
      uri_cb(uri* u,void (uri::*fpt)(char const* a, char const* b))
	:m_uri(u),m_fpt(fpt){}
      uri_cb(const uri_cb& uc)
	: m_uri(uc.m_uri),m_fpt(uc.m_fpt){}
      ~uri_cb(){}
      void operator()(char const* a, char const* b)const{
	(*m_uri.*m_fpt)(a,b);
      }
    private:
      uri *m_uri;
      void (uri::*m_fpt)(char const* a, char const* b);
    };
  private:
    struct uri_grammar : public boost::spirit::grammar<uri_grammar>
    {
    public:
      uri_grammar(uri*const u)
	: m_u(u){}

      template <typename ScannerT>
      struct definition
      {
	typedef boost::spirit::rule<ScannerT> rule_t;

	rule_t URI_reference, absoluteURI, relativeURI, hier_part,
	  opaque_part, uric_no_slash, net_path,abs_path, rel_path,
	  rel_segment, scheme, authority, reg_name, server,userinfo,
	  hostport, host, hostname, domainlabel, toplabel, IPv4address,
	  port, path,  path_segments, segment, param, pchar, query,
	  fragment,uric, reserved, unreserved, mark, escaped, hex,
	  alphanum, alpha,lowalpha, upalpha, digit, dot, e_snum;


	definition(uri_grammar const& ug)
	{
	  // Tokens
	  dot           = boost::spirit::ch_p('.');
	  alpha         = boost::spirit::alpha_p;
	  digit         = boost::spirit::digit_p;
	  alphanum      = boost::spirit::alnum_p;
	  hex           = (digit) | (boost::spirit::ch_p('A') | 'B' | 'C' | 'D' | 'E' | 'F' |
				     'a' | 'b' | 'c' | 'd' | 'e' | 'f');
	  escaped       = boost::spirit::ch_p('%') >> hex >> hex;
	  mark          = (boost::spirit::str_p("'") | (boost::spirit::ch_p('-') | '_' | '.' | '!' | '~' 
					 | '*' | '(' | ')'));
	  unreserved    = alphanum | mark;
	  reserved      = (boost::spirit::ch_p(';') | '/' | '?' | ':' | '@' | '&' | '=' | '+' |
			   '$' | ',');
	  uric_no_slash = unreserved | escaped | (boost::spirit::ch_p(';') | '?' | ':' | '@' |
						  '&' | '=' | '+' | '$' | ',');
	  uric          = reserved | unreserved | escaped;
	  pchar         = unreserved | escaped |
	    (boost::spirit::ch_p(':') | '@' | '&' | '=' | '+' | '$' | ',');

	  // Elements
	  query         = (*uric)[uri_cb(ug.m_u,&uri::on_query)];
	  fragment      = (*uric)[uri_cb(ug.m_u,&uri::on_fragment)];
	  param         = (*pchar)[uri_cb(ug.m_u,&uri::on_param)];
	  segment       = (*pchar >> *( boost::spirit::ch_p(';') >> param ))[uri_cb(ug.m_u,&uri::on_segment)];
	  path_segments = (segment >> *( boost::spirit::ch_p('/') >> segment ))[uri_cb(ug.m_u,&uri::on_path_segments)];
	  abs_path      = (boost::spirit::ch_p('/') >> path_segments)[uri_cb(ug.m_u,&uri::on_abs_path)];
	  path          = (!( abs_path | opaque_part ))[uri_cb(ug.m_u,&uri::on_path)];
	  port          = (*digit)[uri_cb(ug.m_u,&uri::on_port)];
	  IPv4address   = (+digit >> dot >> +digit >> dot >> +digit >> dot >> +digit)[uri_cb(ug.m_u,&uri::on_ipv4)];
	  // Here the syntax is revised from the spec.
	  // Because spirit counts on sequential-and characters being different
	  toplabel      = (+alpha|((alpha >> *( alphanum | boost::spirit::ch_p('-') ) >> alphanum)))[uri_cb(ug.m_u,&uri::on_toplabel)];
	  domainlabel   = (+alphanum |  alphanum >> *( alphanum | boost::spirit::ch_p('-'))  >> alphanum)[uri_cb(ug.m_u,&uri::on_domainlabel)];
	  hostname      = (*( domainlabel >> boost::spirit::ch_p('.') ) >> toplabel >> !( boost::spirit::ch_p('.') ))[uri_cb(ug.m_u,&uri::on_hostname)];
	  host          = (hostname | IPv4address)[uri_cb(ug.m_u,&uri::on_host)];
	  hostport      = (host >> !( boost::spirit::ch_p(':') >> port ))[uri_cb(ug.m_u,&uri::on_hostport)];
	  userinfo      = (*( unreserved | escaped |
			      (boost::spirit::ch_p(';') | ':' | '&' | '=' | '+' | '$' | ',' )))[uri_cb(ug.m_u,&uri::on_userinfo)];
	  server        = (!( !( userinfo >> boost::spirit::ch_p('@') ) >> hostport ))[uri_cb(ug.m_u,&uri::on_server)];
	  reg_name      = (+( unreserved | escaped | (boost::spirit::ch_p('$') | ',' | ';' | ':' | '@' | '&' | '=' | '+' )))[uri_cb(ug.m_u,&uri::on_reg_name)];
	  authority     = (server | reg_name)[uri_cb(ug.m_u,&uri::on_authority)];
	  scheme        = (alpha >> *( alphanum | (boost::spirit::ch_p('+') | '-' | '.') ))[uri_cb(ug.m_u,&uri::on_scheme)];
	  rel_segment   = (+( unreserved | escaped | (boost::spirit::ch_p(';') | '@' | '&' | '=' | '+' | '$' | ',' )))[uri_cb(ug.m_u,&uri::on_rel_segment)];
	  rel_path      = (rel_segment >> !( abs_path ))[uri_cb(ug.m_u,&uri::on_rel_path)];
	  net_path      = (boost::spirit::str_p("//") >> authority >> !( abs_path ))[uri_cb(ug.m_u,&uri::on_net_path)];
	  opaque_part   = (uric_no_slash >> *uric)[uri_cb(ug.m_u,&uri::on_opaque_part)];
	  hier_part     = (( net_path | abs_path ) >> !( boost::spirit::ch_p('?') >> query ))[uri_cb(ug.m_u,&uri::on_hier_part)];
	  relativeURI   = (( net_path | abs_path | rel_path ) >> !( boost::spirit::ch_p('?') >> query ))[uri_cb(ug.m_u,&uri::on_relative_uri)];
	  absoluteURI   = (scheme >> boost::spirit::ch_p(':') >> ( hier_part | opaque_part ))[uri_cb(ug.m_u,&uri::on_absolute_uri)];
	  URI_reference = !(absoluteURI | relativeURI) 
	    >> !( boost::spirit::ch_p('#') >> fragment );
	};
	
	rule_t const& start() const { return URI_reference; }
      };
    private:
      uri*const m_u;      
      
    };
  private:
    void init(uri_grammar& ug);

  };
};
#endif
