/* ====================================================================
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2003 Matt Carter. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Apache" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */

#ifndef MTX_DECIMAL_HPP
#define MTX_DECIMAL_HPP

#include <string>
#include "oapr/exception.hpp"
#include "oapr/errno.hpp"
#include "oapx/math.hpp"
#include <ostream>

namespace oapr {

template<typename int_t, unsigned int decimal_places = 2, typename char_t = char> class decimal
{
  public:
    decimal(): m_value(0) {}
//    decimal(const int_t& src): m_value(src.m_value) {}
    decimal(const decimal& src) { m_value = src.m_value; }
    decimal(const char * const str_value) { MTX_CHECK(parse(str_value)); }
    decimal(const std::string& str_value) { MTX_CHECK(parse(str_value.c_str())); }
    ~decimal() {}

    inline int_t scalar() const { return power(10u, decimal_places); }

    oapr_status_t parse(const char_t * const str)
    {
      // TODO: Handle strtol() out of range error with return MTX_DECIMAL_RANGE_ERROR;
      // TODO: If type supplied is long long call strtoll() instead. Hmm..
      char_t *endptr = NULL;
      int_t dec = strtol(str, &endptr, 10)*scalar();
      bool neg = (dec<0);
//      cerr << "Input: " << str << ", strtol=" << dec << endl;
      if(endptr == str) while(/* *endptr && <-- not needed, isspace doesn't match \0 */ isspace(*endptr)) ++endptr; // Skip leading whitespace if strtol parsed no digits
      if(*endptr == '.') { // We've got a fraction to disassemble
        ++endptr;
        unsigned int index = decimal_places;
        char_t c = *endptr;
//        cerr << "Character (1st): " << c << ", dec=" << dec << endl;
        while(c && index && (c >= '0') && (c <= '9')) {
          // Loop until end of string, end of expected decimal places or first non-digit character
//          cerr << "Character: " << c << ", index: " << index << ", : " << ", dec=" << dec << endl;
          int_t frac;
          if(index>1) frac = power(10u, index-1)*(c-'0'); else frac = c-'0';
          if(neg) dec -= frac; else dec += frac;
          --index;
          ++endptr;
          c = *endptr;
        }
      }
      m_value = dec;
      return MTX_SUCCESS;
    }
    
    friend std::ostream& operator<< <decimal, int_t, decimal_places>(std::ostream&, const decimal&);
    // TODO: friend std::istream& operator>> <decimal, int_t, decimal_places>(std::istream&, decimal&);
    
    inline decimal operator*(int_t v) { return m_value*v.m_value; }
    inline decimal operator*(decimal v) { return m_value*v.m_value/scalar(); }
    inline decimal operator/(decimal v) { return m_value*scalar()/v.m_value; }
    inline decimal operator%(decimal v) { return m_value*scalar()%v.m_value; }
    inline decimal operator+(decimal v) { return decimal(m_value+v.m_value); }
    inline decimal operator-(decimal v) { return m_value-v.m_value; }
    inline int operator==(decimal v) { return m_value == v.m_value; }
    inline int operator!=(decimal v) { return m_value != v.m_value; }

    friend bool operator< <>(decimal, decimal);
    friend bool operator> <>(decimal, decimal);
    friend bool operator<= <>(decimal, decimal);
    friend bool operator>= <>(decimal, decimal);

    /**
     * Change sign of amount (from - to + or vice-versa)
     */
    inline void neg() { m_value = -m_value; }
    decimal& operator=(const decimal& src) { m_value = src.m_value; return *this; }
    decimal& operator=(const char *src) { MTX_CHECK(parse(src)); return *this; }
    decimal& operator=(const std::string& src) { MTX_CHECK(parse(src.c_str())); return *this; }
    decimal& operator+=(decimal v) { m_value += v.m_value; return *this; }
    decimal& operator-=(decimal v) { m_value -= v.m_value; return *this; }
    decimal& operator*=(decimal v) { m_value = m_value*v.m_value/scalar(); return *this; }
    decimal& operator/=(decimal v) { m_value = m_value*scalar()/v.m_value; return *this; }
    decimal& operator%=(decimal v) { m_value = m_value*scalar()%v.m_value; return *this; }
    
    inline int_t trunc() const { return m_value / scalar(); }
    inline int_t frac() const { return m_value % scalar(); }
    inline int_t pennies() const { return m_value; }
    
  private:
    int_t m_value;
    decimal(int_t value): m_value(value) { } // Initialize with value
};

template<class decimal_T> inline bool operator<(decimal_T a, decimal_T b) { return a.m_value < b.m_value; }
template<class decimal_T> inline bool operator>(decimal_T a, decimal_T b) { return a.m_value > b.m_value; }
template<class decimal_T> inline bool operator<=(decimal_T a, decimal_T b) { return a.m_value <= b.m_value; }
template<class decimal_T> inline bool operator>=(decimal_T a, decimal_T b) { return a.m_value >= b.m_value; }

template<class decimal_T, typename int_t, unsigned int decimal_places> std::ostream& operator<<(std::ostream& o, const decimal_T& obj)
{
  o << obj.trunc();
  int_t fr = obj.frac();
  if(fr) {
    o << '.';
    char oldfill = o.fill('0');
    int oldw = o.width(decimal_places);
    o << fr;
    o.width(oldw);
    o.fill(oldfill);
  }
  return o;
}

typedef decimal<long int,2> money;
//typedef decimal<int,4> financial;

}

#endif
