/* ====================================================================
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2003 Justin Kirby. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Apache" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */
#ifndef OAPX_OAPR_IO_MGR_HPP
#define OAPX_OAPR_IO_MGR_HPP
#include "oapr/io.hpp"
#include "oapx/syncbuffer.hpp"
#include "oapr/thread.hpp"
#include "oapr/time.hpp"
#include "oapu/logstream.hpp"
namespace oapr
{
  template <typename char_t, int sz_t>
  class io_mgr_thread : public thread_class
  {
  public:
    typedef sync_buffer<char_t,sz_t> buff_type;
  public:
    io_mgr_thread(basic_io*const io,
		  buff_type*const in,
		  buff_type*const out)
      : m_io(io)
      , m_in(in)
      , m_out(out)
    {
    }
    ~io_mgr_thread()
    {
    }
  protected:
    oapr_status_t execute()
    {
      oapr::logstrm<<"starting\n";
      char bo[sz_t], bi[sz_t];
      std::memset(bo,0,sz_t);
      std::memset(bi,0,sz_t);
      try{
	bool connected=m_io->is_open();
	do{
	  /* we do a read->write first*/
	  if(m_out->avail_r()){
	    oapr::logstrm<<"reading buffer\n";	  
	    int rcv=m_out->read(bo,sz_t);
	    if(rcv>0){
	      int snt(0),rest(rcv);
	      do{
		oapr::logstrm<<"writing to device\n";
		snt+=m_io->write(bo+snt,rest);
		rest=rcv-snt;
	      }while(rest>0);
	    }
	  }

	  /* do we have space to write?*/
	  int isz=m_in->avail_w();
	  if(isz>0){
	    int ravl= isz<sz_t ? isz : sz_t;
	    oapr::logstrm<<"reading from device\n";
	    int rd=m_io->read(bi,ravl);
	    if(rd>0){
	      oapr::logstrm<<"writing to buffer\n";
	      m_in->write(bi,rd);
	    }
	  }
	  oapr::sleep(10);
	  connected=m_io->is_open();

	}while(connected);
	oapr::logstrm<<"exiting\n";
      }catch(...){
	return APR_EGENERAL;
      }
      return APR_SUCCESS;
    }
  private:
    basic_io*const m_io;
    buff_type*const m_in;
    buff_type*const m_out;
  };


  template<typename char_t=char, int sz_t=255>
  class io_manager : public basic_io
  {
  public:
    typedef char_t char_type;
    typedef sync_buffer<char_t,sz_t> buff_type;
    typedef io_mgr_thread<char_t,sz_t> thread_type;
    typedef pool_mgr::pool_t pool_t;
    typedef pool_mgr::pool_ptr pool_ptr;
  public:
    io_manager(const pool_ptr& p)
      : basic_io(p)
	,m_out()
	,m_in()
	, m_io(0)
	, m_thread(0)
    {
    }
    ~io_manager()
    {
      if(m_thread){
	delete m_thread;
      }
    }
    
    void init(basic_io* io, bool now=true)
    {
      /**@todo put a check for already init*/
      m_io=io;
      m_thread = new thread_type(m_io,&m_in,&m_out);
      m_thread->start();
      if(now){
	while(!m_thread->active()){
	  oapr::sleep(1);
	}
      }
    }
    oapr_status_t close()
    {
      if(m_io){
	OAPR_CHECK(m_io->close());
	m_thread->join();
	return APR_SUCCESS;
      }
      else return 0;
    }
    bool eof()
    {
      if(m_io)return m_io->eof();
      else return 0;
    }
    bool is_open()const
    {
      if(m_io)return m_io->is_open();
      else return 0;
    }
    apr_size_t write(const void* buff,
		     apr_size_t toWrite,
		     bool full=false){
      return m_out.write(static_cast<const char*>(buff),toWrite);
    }
    apr_size_t read( void* buff,
		     apr_size_t toRead,
		     bool full=false)const{
      return m_in.read(static_cast<char*>(buff),toRead);
    }
    int size_read(){return m_in.avail_r();}
    int size_write(){return m_out.avail_w();}
    void wait_read(){m_in.wait_r();}

    oapr_status_t flush()
    {
      return APR_ENOTIMPL;
    }


  private:
    buff_type m_out;
    buff_type m_in;
    basic_io *m_io;
    thread_type *m_thread;
  };

 
};
#endif
